import 'package:json_annotation/json_annotation.dart';
import 'package:morph_dart/src/enums/record.dart';
import 'package:morph_dart/src/models/auth/auth.dart';

part 'record.g.dart';

///==================================================================
/// Schema
///==================================================================
abstract class FilterConditionType {
  const FilterConditionType();

  Map<String, dynamic> toJson();
}

@JsonSerializable()
class FilterConditionUnit implements FilterConditionType {
  final String property;
  final String operator;
  final String value;

  const FilterConditionUnit({
    required this.property,
    required this.operator,
    required this.value,
  });

  @override
  Map<String, dynamic> toJson() => _$FilterConditionUnitToJson(this);

  factory FilterConditionUnit.fromJson(Map<String, dynamic> json) =>
      _$FilterConditionUnitFromJson(json);
}

class FilterCondition implements FilterConditionType {
  final List<FilterConditionType>? and;
  final List<FilterConditionType>? or;

  const FilterCondition({
    this.and,
    this.or,
  });

  @override
  Map<String, dynamic> toJson() => converter.toJson(this);

  factory FilterCondition.fromJson(Map<String, dynamic> json) =>
      converter.fromJson(json);

  static const converter = FilterConditionJsonConverter();
}

class FilterConditionJsonConverter
    implements JsonConverter<FilterCondition, Map<String, dynamic>> {
  const FilterConditionJsonConverter();

  @override
  FilterCondition fromJson(Map<String, dynamic> json) {
    final keys = json.keys;
    if (keys.length != 1) {
      throw Exception(
        "Only one of the 'and' or 'or' operator must be specified.",
      );
    }
    final key = keys.first;
    if (key != 'and' || key != 'or') {
      throw Exception(
        "Only one of the 'and' or 'or' operator must be specified.",
      );
    }

    final List<FilterConditionType> filterChildren = [];

    for (final filter in json[key] as List<dynamic>) {
      if ((filter as Map<String, dynamic>).containsKey('property')) {
        filterChildren.add(FilterConditionUnit.fromJson(filter));
      } else {
        filterChildren.add(FilterCondition.fromJson(filter));
      }
    }

    return FilterCondition(
      and: key == 'and' ? filterChildren : null,
      or: key == 'or' ? filterChildren : null,
    );
  }

  @override
  Map<String, dynamic> toJson(FilterCondition data) {
    final Map<String, dynamic> json = {};
    final and = data.and;
    final or = data.or;

    if (and != null) {
      final List<Map<String, dynamic>> list = [];
      for (final element in and) {
        list.add(element.toJson());
      }
      json['and'] = list;
    }

    if (or != null) {
      final List<Map<String, dynamic>> list = [];
      for (final element in or) {
        list.add(element.toJson());
      }
      json['or'] = list;
    }

    return json;
  }
}

@JsonSerializable()
class SortCondition {
  final String property;
  final SortConditionDirection direction;

  const SortCondition({
    required this.property,
    required this.direction,
  });

  Map<String, dynamic> toJson() => _$SortConditionToJson(this);

  factory SortCondition.fromJson(Map<String, dynamic> json) =>
      _$SortConditionFromJson(json);
}

@JsonSerializable()
class RecordObject {
  final String id;
  final Map<String, dynamic> fields;
  final UserpoolTeamUserObject? createdBy;
  final String createdAt;
  final String? updateAt;

  const RecordObject({
    required this.id,
    required this.fields,
    this.createdBy,
    required this.createdAt,
    this.updateAt,
  });

  factory RecordObject.fromJson(Map<String, dynamic> json) =>
      _$RecordObjectFromJson(json);
}

///==================================================================
/// Request Body
///==================================================================
@JsonSerializable(explicitToJson: true)
class QueryRecordRequestBody {
  final List<String>? select;
  final FilterCondition? filter;
  final List<SortCondition>? sorts;
  final num? limit;
  final num? skip;

  @JsonKey(includeIfNull: false)
  final String? userpoolTeamId;
  @JsonKey(includeIfNull: false)
  final String? groupId;
  @JsonKey(includeIfNull: false)
  final bool? withNextLayer;

  const QueryRecordRequestBody({
    this.select,
    this.filter,
    this.sorts,
    this.limit,
    this.skip,
    this.userpoolTeamId,
    this.groupId,
    this.withNextLayer,
  });

  Map<String, dynamic> toJson() => _$QueryRecordRequestBodyToJson(this);
}

@JsonSerializable(explicitToJson: true)
class DownloadRecordRequestBody extends QueryRecordRequestBody {
  final DownloadFormat? format;

  const DownloadRecordRequestBody({
    List<String>? select,
    FilterCondition? filter,
    List<SortCondition>? sorts,
    num? limit,
    num? skip,
    String? userpoolTeamId,
    String? groupId,
    bool? withNextLayer,
    this.format,
  }) : super(
          select: select,
          filter: filter,
          sorts: sorts,
          limit: limit,
          skip: skip,
          userpoolTeamId: userpoolTeamId,
          groupId: groupId,
          withNextLayer: withNextLayer,
        );

  @override
  Map<String, dynamic> toJson() => _$DownloadRecordRequestBodyToJson(this);
}

@JsonSerializable()
class CreateRecordRequestBody {
  final Map<String, dynamic> data;
  final String? userpoolTeamId;
  final String? groupId;

  const CreateRecordRequestBody({
    required this.data,
    this.userpoolTeamId,
    this.groupId,
  });

  Map<String, dynamic> toJson() => _$CreateRecordRequestBodyToJson(this);
}

@JsonSerializable()
class UpdateRecordRequestBody {
  final Map<String, dynamic> data;

  const UpdateRecordRequestBody({
    required this.data,
  });

  Map<String, dynamic> toJson() => _$UpdateRecordRequestBodyToJson(this);
}

///==================================================================
/// Response
///==================================================================
@JsonSerializable()
class RecordResponse extends RecordObject {
  const RecordResponse({
    required String id,
    required Map<String, dynamic> fields,
    UserpoolTeamUserObject? createdBy,
    required String createdAt,
    String? updateAt,
  }) : super(
          id: id,
          fields: fields,
          createdBy: createdBy,
          createdAt: createdAt,
          updateAt: updateAt,
        );

  factory RecordResponse.fromJson(Map<String, dynamic> json) =>
      _$RecordResponseFromJson(json);
}

@JsonSerializable()
class RecordListResponse {
  final List<RecordObject> records;
  final num count;

  const RecordListResponse({required this.records, required this.count});

  factory RecordListResponse.fromJson(Map<String, dynamic> json) =>
      _$RecordListResponseFromJson(json);
}

@JsonSerializable()
class DownloadRecordResponse {
  final String url;
  final DownloadFormat format;
  final num count;

  const DownloadRecordResponse({
    required this.url,
    required this.format,
    required this.count,
  });

  factory DownloadRecordResponse.fromJson(Map<String, dynamic> json) =>
      _$DownloadRecordResponseFromJson(json);
}
